
// a simple class for string trees.
// This version keeps the size of each subtree in each node's count_ attribute.
// count_ is to support an O(log N) implementation of Get(k)

class StrTreeC {  // StrTree with count attribute
     
    // We do not provide a way to set the string in an existing node; 
    // Providing such a method is a bad idea for ordered structures (why?)
    class treenode {
	private String data_;
	private int count_;	    // should always be the size of this SUBTREE
	private int multiplicity_;  // optional; # of times this node's data is counted
	private treenode left_, right_;
	public treenode(String d, treenode l, treenode r) {
		data_ = d; 
		left_ = l; 
		right_ = r;
		int lcount, rcount;
		if (left_  == null) lcount = 0; else lcount = left_.count();
		if (right_ == null) rcount = 0; else rcount = right_.count();
		count_ = lcount + rcount + 1; 
	}
	public String data() {return data_;}
	public int   count() {return count_;}
	public int multiplicity() {return multiplicity_;}
	public treenode left()  {return left_;}
	public treenode right() {return right_;}
	public void inccount() {count_ += 1;}
	//public void setcount(int count) {count_ = count;}
	public void incmultiplicity() {multiplicity_ += 1;}
	public void setmultiplicity(int count) {multiplicity_ = count;}
	public void setleft  (treenode t) {left_ = t;} // doesn't update count_!
	public void setright (treenode t) {right_ = t;} // doesn't update count_!
    }

    private treenode root_ = null;

    public void inTraverse() {inTraverse(root_, 0);}

    private void inTraverse(treenode p, int depth) {
    	int indent = 4;
    	if (p==null) return;
    	inTraverse(p.left(), depth+1);
    	String prefix = "";
    	for (int i=0; i<depth*indent; i++) prefix += ' ';
    	System.out.printf("%s%s  (%d)%n", prefix, p.data(), p.count());
    	inTraverse(p.right(), depth+1);
    }

    // must not be called with val == null
    public void insert(String val) {
    	if (root_ == null) {
    		root_ = new treenode(val, null, null); // sets count_, but not multiplicity_
    		return;
    	} else {
    		rinsert(val, root_);
    	}
    }

  // recursive insert: does not get called with t==null
  // returns true if t's count() was increased (so parent's count can also be increased)
    private boolean rinsert(String val, treenode t) {
    	int comp = val.compareTo(t.data()); // = 0 if Strings ==, <0 if val<t.data, etc
    	if (comp == 0 ) {
    		return false;	// This is where you would update multiplicity_
    	} else if (comp < 0) {		// val < t.data()
    	    if (t.left() == null) {
        		t.setleft(new treenode(val, null, null));  // t.left()'s count is now 1
        		t.inccount();
        		return true;
    	    } else {
        		boolean leftchange = rinsert(val, t.left());
        		if (leftchange) t.inccount();
        		return leftchange;
    	    }
    	} else { // val > t.data(); comp > 0
    	    if (t.right() == null) {		
        		t.setright(new treenode(val, null, null));
        		t.inccount();
        		return true;
    	    } else {				// FIX so counts are updated
        		boolean rightchange = rinsert(val, t.right());
        		if (rightchange) t.inccount();
        		return true;
    	    }
    	}
    }

}


